/*
 *  Copyright 2021 Sourcesense Spa
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.sourcesense.joyce.core.configuration.kafka;

import com.fasterxml.jackson.databind.node.ObjectNode;
import io.opentracing.Tracer;
import io.opentracing.contrib.kafka.TracingConsumerInterceptor;
import io.opentracing.contrib.kafka.spring.TracingConsumerFactory;
import io.opentracing.contrib.kafka.spring.TracingKafkaAspect;
import lombok.RequiredArgsConstructor;
import org.apache.kafka.clients.consumer.ConsumerConfig;
import org.apache.kafka.common.serialization.StringDeserializer;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.boot.autoconfigure.condition.ConditionalOnProperty;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.kafka.annotation.EnableKafka;
import org.springframework.kafka.config.ConcurrentKafkaListenerContainerFactory;
import org.springframework.kafka.core.ConsumerFactory;
import org.springframework.kafka.core.DefaultKafkaConsumerFactory;
import org.springframework.kafka.support.serializer.ErrorHandlingDeserializer;
import org.springframework.kafka.support.serializer.JsonDeserializer;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

@ConditionalOnProperty(value = "joyce.kafka.consumer.enabled", havingValue = "true")
@Configuration
@EnableKafka
@RequiredArgsConstructor
public class KafkaConsumerConfig {
		final private Tracer tracer;

    @Value("${joyce.kafka.bootstrapAddress}")
    String bootstrapAddress;

    @Value("${joyce.kafka.consumer.groupId:joyce-consumer-local}")
    String groupId;

		@Value("${joyce.kafka.consumer.batch-size:100}")
		Integer batchSize;

    @Bean
    public ConsumerFactory<String, ObjectNode> consumerFactory() {
        Map<String, Object> props = new HashMap<>();
        props.put(
                ConsumerConfig.BOOTSTRAP_SERVERS_CONFIG,
                bootstrapAddress);
        props.put(
                ConsumerConfig.GROUP_ID_CONFIG,
                groupId);

        props.put(ConsumerConfig.VALUE_DESERIALIZER_CLASS_CONFIG, ErrorHandlingDeserializer.class);
        props.put(ConsumerConfig.KEY_DESERIALIZER_CLASS_CONFIG, ErrorHandlingDeserializer.class);
        props.put(ErrorHandlingDeserializer.KEY_DESERIALIZER_CLASS, StringDeserializer.class);

        props.put(ErrorHandlingDeserializer.VALUE_DESERIALIZER_CLASS, JsonDeserializer.class.getName());
        props.put(JsonDeserializer.VALUE_DEFAULT_TYPE, "com.fasterxml.jackson.databind.node.ObjectNode");
				props.put(ConsumerConfig.MAX_POLL_INTERVAL_MS_CONFIG, 900000);
				props.put(ConsumerConfig.MAX_POLL_RECORDS_CONFIG, batchSize);
				props.put(ConsumerConfig.HEARTBEAT_INTERVAL_MS_CONFIG, 3000);

//				props.put(ConsumerConfig.INTERCEPTOR_CLASSES_CONFIG, Collections.singletonList(TracingConsumerInterceptor.class));


//			return new DefaultKafkaConsumerFactory<>(props);
			return new TracingConsumerFactory<>(new DefaultKafkaConsumerFactory<>(props), tracer);

		}

		@Bean
		public TracingKafkaAspect tracingKafkaAspect() {
			return new TracingKafkaAspect(tracer);
		}

    @Bean
    public ConcurrentKafkaListenerContainerFactory<String, ObjectNode>
    kafkaListenerContainerFactory() {
        ConcurrentKafkaListenerContainerFactory<String, ObjectNode> factory =
                new ConcurrentKafkaListenerContainerFactory<>();
        factory.setConsumerFactory(consumerFactory());
        return factory;
    }

	@Bean
	@ConditionalOnProperty(value = "joyce.kafka.consumer.batch", havingValue = "true")
	public ConcurrentKafkaListenerContainerFactory<String, ObjectNode>
	kafkaBatchListenerContainerFactory() {
		ConcurrentKafkaListenerContainerFactory<String, ObjectNode> factory =
				new ConcurrentKafkaListenerContainerFactory<>();
		factory.setConsumerFactory(consumerFactory());
		factory.setBatchListener(true);
		return factory;
	}
}
